#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that creates, update, get and delete trigger for Data Loss Prevention actions.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dlp import (
    CloudDLPCreateJobTriggerOperator,
    CloudDLPDeleteJobTriggerOperator,
    CloudDLPGetDLPJobTriggerOperator,
    CloudDLPListJobTriggersOperator,
    CloudDLPUpdateJobTriggerOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "dlp_job_trigger"
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

JOB_TRIGGER = {
    "inspect_job": {
        "storage_config": {
            "datastore_options": {"partition_id": {"project_id": PROJECT_ID}, "kind": {"name": "test"}}
        }
    },
    "triggers": [{"schedule": {"recurrence_period_duration": {"seconds": 60 * 60 * 24}}}],
    "status": "HEALTHY",
}

TRIGGER_ID = f"trigger_{ENV_ID}"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["dlp", "example"],
) as dag:
    # [START howto_operator_dlp_create_job_trigger]
    create_trigger = CloudDLPCreateJobTriggerOperator(
        project_id=PROJECT_ID,
        job_trigger=JOB_TRIGGER,
        trigger_id=TRIGGER_ID,
        task_id="create_trigger",
    )
    # [END howto_operator_dlp_create_job_trigger]

    list_triggers = CloudDLPListJobTriggersOperator(task_id="list_triggers", project_id=PROJECT_ID)

    get_trigger = CloudDLPGetDLPJobTriggerOperator(
        task_id="get_trigger", project_id=PROJECT_ID, job_trigger_id=TRIGGER_ID
    )

    JOB_TRIGGER["triggers"] = [{"schedule": {"recurrence_period_duration": {"seconds": 2 * 60 * 60 * 24}}}]

    # [START howto_operator_dlp_update_job_trigger]
    update_trigger = CloudDLPUpdateJobTriggerOperator(
        project_id=PROJECT_ID,
        job_trigger_id=TRIGGER_ID,
        job_trigger=JOB_TRIGGER,
        task_id="update_info_type",
    )
    # [END howto_operator_dlp_update_job_trigger]

    # [START howto_operator_dlp_delete_job_trigger]
    delete_trigger = CloudDLPDeleteJobTriggerOperator(
        project_id=PROJECT_ID, job_trigger_id=TRIGGER_ID, task_id="delete_info_type"
    )
    # [END howto_operator_dlp_delete_job_trigger]
    delete_trigger.trigger_rule = TriggerRule.ALL_DONE

    (create_trigger >> list_triggers >> get_trigger >> update_trigger >> delete_trigger)

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Memorystore Memcached service.

This DAG relies on the following OS environment variables

* AIRFLOW__API__GOOGLE_KEY_PATH - Path to service account key file. Note, you can skip this variable if you
  run this DAG in a Composer environment.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.protobuf.field_mask_pb2 import FieldMask

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.cloud_memorystore import (
    CloudMemorystoreMemcachedApplyParametersOperator,
    CloudMemorystoreMemcachedCreateInstanceOperator,
    CloudMemorystoreMemcachedDeleteInstanceOperator,
    CloudMemorystoreMemcachedGetInstanceOperator,
    CloudMemorystoreMemcachedListInstancesOperator,
    CloudMemorystoreMemcachedUpdateInstanceOperator,
    CloudMemorystoreMemcachedUpdateParametersOperator,
)
from airflow.providers.standard.operators.bash import BashOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "google_project_id")

DAG_ID = "cloud_memorystore_memcached"

MEMORYSTORE_MEMCACHED_INSTANCE_NAME = f"memcached-{ENV_ID}-1"
LOCATION = "europe-north1"

# [START howto_operator_memcached_instance]
MEMCACHED_INSTANCE = {
    "name": "",
    "node_count": 1,
    "node_config": {"cpu_count": 1, "memory_size_mb": 1024},
    "zones": [LOCATION + "-a"],
}
# [END howto_operator_memcached_instance]

IP_RANGE_NAME = f"ip-range-{DAG_ID}".replace("_", "-")
NETWORK = "default"
CREATE_PRIVATE_CONNECTION_CMD = f"""
if [ $AIRFLOW__API__GOOGLE_KEY_PATH ]; then \
 gcloud auth activate-service-account --key-file=$AIRFLOW__API__GOOGLE_KEY_PATH; \
fi;
if [[ $(gcloud compute addresses list --project={PROJECT_ID} --filter="name=('{IP_RANGE_NAME}')") ]]; then \
  echo "The IP range '{IP_RANGE_NAME}' already exists in the project '{PROJECT_ID}'."; \
else \
  echo "  Creating IP range..."; \
  gcloud compute addresses create "{IP_RANGE_NAME}" \
    --global \
    --purpose=VPC_PEERING \
    --prefix-length=16 \
    --description="IP range for Memorystore system tests" \
    --network={NETWORK} \
    --project={PROJECT_ID}; \
  echo "Done."; \
fi;
if [[ $(gcloud services vpc-peerings list --network={NETWORK} --project={PROJECT_ID}) ]]; then \
  echo "The private connection already exists in the project '{PROJECT_ID}'."; \
else \
  echo "  Creating private connection..."; \
  gcloud services vpc-peerings connect \
    --service=servicenetworking.googleapis.com \
    --ranges={IP_RANGE_NAME} \
    --network={NETWORK} \
    --project={PROJECT_ID}; \
  echo "Done."; \
fi;
if [[ $(gcloud services vpc-peerings list \
        --network={NETWORK} \
        --project={PROJECT_ID} \
        --format="value(reservedPeeringRanges)" | grep {IP_RANGE_NAME}) ]]; then \
  echo "Private service connection configured."; \
else \
  echo "  Updating service private connection..."; \
  gcloud services vpc-peerings update \
    --service=servicenetworking.googleapis.com \
    --ranges={IP_RANGE_NAME} \
    --network={NETWORK} \
    --project={PROJECT_ID} \
    --force; \
fi;
"""

with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example"],
) as dag:
    create_private_service_connection = BashOperator(
        task_id="create_private_service_connection",
        bash_command=CREATE_PRIVATE_CONNECTION_CMD,
    )

    # [START howto_operator_create_instance_memcached]
    create_memcached_instance = CloudMemorystoreMemcachedCreateInstanceOperator(
        task_id="create-instance",
        location=LOCATION,
        instance_id=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        instance=MEMCACHED_INSTANCE,
        project_id=PROJECT_ID,
    )
    # [END howto_operator_create_instance_memcached]

    # [START howto_operator_delete_instance_memcached]
    delete_memcached_instance = CloudMemorystoreMemcachedDeleteInstanceOperator(
        task_id="delete-instance",
        location=LOCATION,
        instance=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        project_id=PROJECT_ID,
    )
    # [END howto_operator_delete_instance_memcached]
    delete_memcached_instance.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_get_instance_memcached]
    get_memcached_instance = CloudMemorystoreMemcachedGetInstanceOperator(
        task_id="get-instance",
        location=LOCATION,
        instance=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        project_id=PROJECT_ID,
    )
    # [END howto_operator_get_instance_memcached]

    # [START howto_operator_list_instances_memcached]
    list_memcached_instances = CloudMemorystoreMemcachedListInstancesOperator(
        task_id="list-instances", location="-", project_id=PROJECT_ID
    )
    # [END howto_operator_list_instances_memcached]

    # [START howto_operator_update_instance_memcached]
    update_memcached_instance = CloudMemorystoreMemcachedUpdateInstanceOperator(
        task_id="update-instance",
        location=LOCATION,
        instance_id=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        project_id=PROJECT_ID,
        update_mask=FieldMask(paths=["node_count"]),
        instance={"node_count": 2},  # 2
    )
    # [END howto_operator_update_instance_memcached]

    # [START howto_operator_update_and_apply_parameters_memcached]
    update_memcached_parameters = CloudMemorystoreMemcachedUpdateParametersOperator(
        task_id="update-parameters",
        location=LOCATION,
        instance_id=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        project_id=PROJECT_ID,
        update_mask={"paths": ["params"]},
        parameters={"params": {"protocol": "ascii", "hash_algorithm": "jenkins"}},
    )

    apply_memcached_parameters = CloudMemorystoreMemcachedApplyParametersOperator(
        task_id="apply-parameters",
        location=LOCATION,
        instance_id=MEMORYSTORE_MEMCACHED_INSTANCE_NAME,
        project_id=PROJECT_ID,
        node_ids=["node-a-1"],
        apply_all=False,
    )
    # [END howto_operator_update_and_apply_parameters_memcached]

    (
        create_private_service_connection
        >> create_memcached_instance
        >> get_memcached_instance
        >> list_memcached_instances
        >> update_memcached_instance
        >> update_memcached_parameters
        >> apply_memcached_parameters
        >> delete_memcached_instance
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from uuid import UUID

import httplib2
import pytest
from google.api_core.exceptions import AlreadyExists, GoogleAPICallError
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.exceptions import NotFound
from google.cloud.pubsub_v1.types import PublisherOptions, ReceivedMessage
from googleapiclient.errors import HttpError

from airflow.providers.google.cloud.hooks.pubsub import PubSubAsyncHook, PubSubException, PubSubHook
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.version import version

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
PUBSUB_STRING = "airflow.providers.google.cloud.hooks.pubsub.{}"

EMPTY_CONTENT = b""
TEST_PROJECT = "test-project"
TEST_TOPIC = "test-topic"
TEST_SUBSCRIPTION = "test-subscription"
TEST_UUID = UUID("cf4a56d2-8101-4217-b027-2af6216feb48")
TEST_MESSAGES = [
    {"data": b"Hello, World!", "attributes": {"type": "greeting"}},
    {"data": b"Knock, knock"},
    {"attributes": {"foo": ""}},
]

EXPANDED_TOPIC = f"projects/{TEST_PROJECT}/topics/{TEST_TOPIC}"
EXPANDED_SUBSCRIPTION = f"projects/{TEST_PROJECT}/subscriptions/{TEST_SUBSCRIPTION}"
LABELS = {"airflow-version": "v" + version.replace(".", "-").replace("+", "-")}


def mock_init(
    self,
    gcp_conn_id,
    impersonation_chain=None,
):
    pass


def _generate_messages(count) -> list[ReceivedMessage]:
    return [
        ReceivedMessage(
            ack_id=str(i),
            message={
                "data": f"Message {i}".encode(),
                "attributes": {"type": "generated message"},
            },
        )
        for i in range(1, count + 1)
    ]


class TestPubSubHook:
    def setup_method(self):
        with mock.patch(BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_init):
            self.pubsub_hook = PubSubHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PublisherClient")
    def test_publisher_client_creation(self, mock_client, mock_get_creds):
        assert self.pubsub_hook._client is None
        result = self.pubsub_hook.get_conn()

        mock_client.assert_called_once_with(
            credentials=mock_get_creds.return_value,
            client_info=CLIENT_INFO,
            publisher_options=PublisherOptions(enable_message_ordering=False),
        )
        assert mock_client.return_value == result
        assert self.pubsub_hook._client == result

    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.SubscriberClient")
    def test_subscriber_client_creation(self, mock_client, mock_get_creds):
        assert self.pubsub_hook._client is None
        result = self.pubsub_hook.subscriber_client
        mock_client.assert_called_once_with(credentials=mock_get_creds.return_value, client_info=CLIENT_INFO)
        assert mock_client.return_value == result

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_create_nonexistent_topic(self, mock_service):
        create_method = mock_service.return_value.create_topic
        self.pubsub_hook.create_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC)
        create_method.assert_called_once_with(
            request=dict(
                name=EXPANDED_TOPIC,
                labels=LABELS,
                message_storage_policy=None,
                kms_key_name=None,
                schema_settings=None,
                message_retention_duration=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_delete_topic(self, mock_service):
        delete_method = mock_service.return_value.delete_topic
        self.pubsub_hook.delete_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC)
        delete_method.assert_called_once_with(
            request=dict(topic=EXPANDED_TOPIC), retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_delete_nonexisting_topic_failifnotexists(self, mock_service):
        mock_service.return_value.delete_topic.side_effect = NotFound(
            f"Topic does not exists: {EXPANDED_TOPIC}"
        )
        with pytest.raises(PubSubException) as ctx:
            self.pubsub_hook.delete_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC, fail_if_not_exists=True)

        assert str(ctx.value) == f"Topic does not exist: {EXPANDED_TOPIC}"

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_delete_topic_api_call_error(self, mock_service):
        mock_service.return_value.delete_topic.side_effect = GoogleAPICallError(
            f"Error deleting topic: {EXPANDED_TOPIC}"
        )
        with pytest.raises(PubSubException):
            self.pubsub_hook.delete_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC, fail_if_not_exists=True)

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_create_preexisting_topic_failifexists(self, mock_service):
        mock_service.return_value.create_topic.side_effect = AlreadyExists(
            f"Topic already exists: {TEST_TOPIC}"
        )
        with pytest.raises(PubSubException) as ctx:
            self.pubsub_hook.create_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC, fail_if_exists=True)
        assert str(ctx.value) == f"Topic already exists: {TEST_TOPIC}"

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_create_preexisting_topic_nofailifexists(self, mock_service):
        mock_service.return_value.create_topic.side_effect = AlreadyExists(
            f"Topic already exists: {EXPANDED_TOPIC}"
        )
        self.pubsub_hook.create_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC)

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_create_topic_api_call_error(self, mock_service):
        mock_service.return_value.create_topic.side_effect = GoogleAPICallError(
            f"Error creating topic: {TEST_TOPIC}"
        )
        with pytest.raises(PubSubException):
            self.pubsub_hook.create_topic(project_id=TEST_PROJECT, topic=TEST_TOPIC, fail_if_exists=True)

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_nonexistent_subscription(self, mock_service):
        create_method = mock_service.create_subscription

        response = self.pubsub_hook.create_subscription(
            project_id=TEST_PROJECT, topic=TEST_TOPIC, subscription=TEST_SUBSCRIPTION
        )
        create_method.assert_called_once_with(
            request=dict(
                name=EXPANDED_SUBSCRIPTION,
                topic=EXPANDED_TOPIC,
                push_config=None,
                ack_deadline_seconds=10,
                retain_acked_messages=None,
                message_retention_duration=None,
                labels=LABELS,
                enable_message_ordering=False,
                expiration_policy=None,
                filter=None,
                dead_letter_policy=None,
                retry_policy=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert response == TEST_SUBSCRIPTION

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_different_project_topic(self, mock_service):
        create_method = mock_service.create_subscription
        response = self.pubsub_hook.create_subscription(
            project_id=TEST_PROJECT,
            topic=TEST_TOPIC,
            subscription=TEST_SUBSCRIPTION,
            subscription_project_id="a-different-project",
        )
        expected_subscription = f"projects/a-different-project/subscriptions/{TEST_SUBSCRIPTION}"
        create_method.assert_called_once_with(
            request=dict(
                name=expected_subscription,
                topic=EXPANDED_TOPIC,
                push_config=None,
                ack_deadline_seconds=10,
                retain_acked_messages=None,
                message_retention_duration=None,
                labels=LABELS,
                enable_message_ordering=False,
                expiration_policy=None,
                filter=None,
                dead_letter_policy=None,
                retry_policy=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        assert response == TEST_SUBSCRIPTION

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_delete_subscription(self, mock_service):
        self.pubsub_hook.delete_subscription(project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION)
        delete_method = mock_service.delete_subscription
        delete_method.assert_called_once_with(
            request=dict(subscription=EXPANDED_SUBSCRIPTION), retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_delete_nonexisting_subscription_failifnotexists(self, mock_service):
        mock_service.delete_subscription.side_effect = NotFound(
            f"Subscription does not exists: {EXPANDED_SUBSCRIPTION}"
        )
        with pytest.raises(PubSubException) as ctx:
            self.pubsub_hook.delete_subscription(
                project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, fail_if_not_exists=True
            )
        assert str(ctx.value) == f"Subscription does not exist: {EXPANDED_SUBSCRIPTION}"

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_delete_subscription_api_call_error(self, mock_service):
        mock_service.delete_subscription.side_effect = GoogleAPICallError(
            f"Error deleting subscription {EXPANDED_SUBSCRIPTION}"
        )
        with pytest.raises(PubSubException):
            self.pubsub_hook.delete_subscription(
                project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, fail_if_not_exists=True
            )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    @mock.patch(PUBSUB_STRING.format("uuid4"), new_callable=mock.Mock(return_value=lambda: TEST_UUID))
    def test_create_subscription_without_subscription_name(self, mock_uuid, mock_service):
        create_method = mock_service.create_subscription
        expected_name = EXPANDED_SUBSCRIPTION.replace(TEST_SUBSCRIPTION, f"sub-{TEST_UUID}")

        response = self.pubsub_hook.create_subscription(project_id=TEST_PROJECT, topic=TEST_TOPIC)
        create_method.assert_called_once_with(
            request=dict(
                name=expected_name,
                topic=EXPANDED_TOPIC,
                push_config=None,
                ack_deadline_seconds=10,
                retain_acked_messages=None,
                message_retention_duration=None,
                labels=LABELS,
                enable_message_ordering=False,
                expiration_policy=None,
                filter=None,
                dead_letter_policy=None,
                retry_policy=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert f"sub-{TEST_UUID}" == response

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_with_ack_deadline(self, mock_service):
        create_method = mock_service.create_subscription

        response = self.pubsub_hook.create_subscription(
            project_id=TEST_PROJECT, topic=TEST_TOPIC, subscription=TEST_SUBSCRIPTION, ack_deadline_secs=30
        )
        create_method.assert_called_once_with(
            request=dict(
                name=EXPANDED_SUBSCRIPTION,
                topic=EXPANDED_TOPIC,
                push_config=None,
                ack_deadline_seconds=30,
                retain_acked_messages=None,
                message_retention_duration=None,
                labels=LABELS,
                enable_message_ordering=False,
                expiration_policy=None,
                filter=None,
                dead_letter_policy=None,
                retry_policy=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert response == TEST_SUBSCRIPTION

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_with_filter(self, mock_service):
        create_method = mock_service.create_subscription

        response = self.pubsub_hook.create_subscription(
            project_id=TEST_PROJECT,
            topic=TEST_TOPIC,
            subscription=TEST_SUBSCRIPTION,
            filter_='attributes.domain="com"',
        )
        create_method.assert_called_once_with(
            request=dict(
                name=EXPANDED_SUBSCRIPTION,
                topic=EXPANDED_TOPIC,
                push_config=None,
                ack_deadline_seconds=10,
                retain_acked_messages=None,
                message_retention_duration=None,
                labels=LABELS,
                enable_message_ordering=False,
                expiration_policy=None,
                filter='attributes.domain="com"',
                dead_letter_policy=None,
                retry_policy=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert response == TEST_SUBSCRIPTION

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_failifexists(self, mock_service):
        mock_service.create_subscription.side_effect = AlreadyExists(
            f"Subscription already exists: {EXPANDED_SUBSCRIPTION}"
        )
        with pytest.raises(PubSubException) as ctx:
            self.pubsub_hook.create_subscription(
                project_id=TEST_PROJECT, topic=TEST_TOPIC, subscription=TEST_SUBSCRIPTION, fail_if_exists=True
            )
        assert str(ctx.value) == f"Subscription already exists: {EXPANDED_SUBSCRIPTION}"

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_api_call_error(self, mock_service):
        mock_service.create_subscription.side_effect = GoogleAPICallError(
            f"Error creating subscription {EXPANDED_SUBSCRIPTION}"
        )
        with pytest.raises(PubSubException):
            self.pubsub_hook.create_subscription(
                project_id=TEST_PROJECT, topic=TEST_TOPIC, subscription=TEST_SUBSCRIPTION, fail_if_exists=True
            )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_create_subscription_nofailifexists(self, mock_service):
        mock_service.create_subscription.side_effect = AlreadyExists(
            f"Subscription already exists: {EXPANDED_SUBSCRIPTION}"
        )
        response = self.pubsub_hook.create_subscription(
            project_id=TEST_PROJECT, topic=TEST_TOPIC, subscription=TEST_SUBSCRIPTION
        )
        assert response == TEST_SUBSCRIPTION

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_publish(self, mock_service):
        publish_method = mock_service.return_value.publish

        self.pubsub_hook.publish(project_id=TEST_PROJECT, topic=TEST_TOPIC, messages=TEST_MESSAGES)
        calls = [
            mock.call(topic=EXPANDED_TOPIC, data=message.get("data", b""), **message.get("attributes", {}))
            for message in TEST_MESSAGES
        ]
        mock_calls_result = publish_method.mock_calls
        result_refined = [mock_calls_result[0], mock_calls_result[2], mock_calls_result[4]]
        assert result_refined == calls

    @mock.patch(PUBSUB_STRING.format("PubSubHook.get_conn"))
    def test_publish_api_call_error(self, mock_service):
        publish_method = mock_service.return_value.publish
        publish_method.side_effect = GoogleAPICallError(f"Error publishing to topic {EXPANDED_SUBSCRIPTION}")

        with pytest.raises(PubSubException):
            self.pubsub_hook.publish(project_id=TEST_PROJECT, topic=TEST_TOPIC, messages=TEST_MESSAGES)

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_pull(self, mock_service):
        pull_method = mock_service.pull
        pulled_messages = []
        for i, msg in enumerate(TEST_MESSAGES):
            pulled_messages.append({"ackId": i, "message": msg})
        pull_method.return_value.received_messages = pulled_messages

        response = self.pubsub_hook.pull(
            project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, max_messages=10
        )
        pull_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                max_messages=10,
                return_immediately=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert pulled_messages == response

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_pull_no_messages(self, mock_service):
        pull_method = mock_service.pull
        pull_method.return_value.received_messages = []

        response = self.pubsub_hook.pull(
            project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, max_messages=10
        )
        pull_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                max_messages=10,
                return_immediately=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert response == []

    @pytest.mark.parametrize(
        "exception",
        [
            pytest.param(
                HttpError(resp=httplib2.Response({"status": 404}), content=EMPTY_CONTENT), id="http-error-404"
            ),
            pytest.param(GoogleAPICallError("API Call Error"), id="google-api-call-error"),
        ],
    )
    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_pull_fails_on_exception(self, mock_service, exception):
        pull_method = mock_service.pull
        pull_method.side_effect = exception

        with pytest.raises(PubSubException):
            self.pubsub_hook.pull(project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, max_messages=10)
        pull_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                max_messages=10,
                return_immediately=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_acknowledge_by_ack_ids(self, mock_service):
        ack_method = mock_service.acknowledge

        self.pubsub_hook.acknowledge(
            project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, ack_ids=["1", "2", "3"]
        )
        ack_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                ack_ids=["1", "2", "3"],
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_acknowledge_by_message_objects(self, mock_service):
        ack_method = mock_service.acknowledge

        self.pubsub_hook.acknowledge(
            project_id=TEST_PROJECT,
            subscription=TEST_SUBSCRIPTION,
            messages=_generate_messages(3),
        )
        ack_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                ack_ids=["1", "2", "3"],
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @pytest.mark.parametrize(
        "ack_ids, messages",
        [
            pytest.param(None, None, id="both-empty"),
            pytest.param([1, 2, 3], _generate_messages(3), id="both-provided"),
        ],
    )
    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_acknowledge_fails_on_method_args_validation(self, mock_service, ack_ids, messages):
        ack_method = mock_service.acknowledge

        error_message = r"One and only one of 'ack_ids' and 'messages' arguments have to be provided"
        with pytest.raises(ValueError, match=error_message):
            self.pubsub_hook.acknowledge(
                project_id=TEST_PROJECT,
                subscription=TEST_SUBSCRIPTION,
                ack_ids=ack_ids,
                messages=messages,
            )
        ack_method.assert_not_called()

    @pytest.mark.parametrize(
        "exception",
        [
            pytest.param(
                HttpError(resp=httplib2.Response({"status": 404}), content=EMPTY_CONTENT), id="http-error-404"
            ),
            pytest.param(GoogleAPICallError("API Call Error"), id="google-api-call-error"),
        ],
    )
    @mock.patch(PUBSUB_STRING.format("PubSubHook.subscriber_client"))
    def test_acknowledge_fails_on_exception(self, mock_service, exception):
        ack_method = mock_service.acknowledge
        ack_method.side_effect = exception

        with pytest.raises(PubSubException):
            self.pubsub_hook.acknowledge(
                project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, ack_ids=["1", "2", "3"]
            )
        ack_method.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                ack_ids=["1", "2", "3"],
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @pytest.mark.parametrize(
        "messages",
        [
            [{"data": b"test"}],
            [{"data": b""}],
            [{"data": b"test", "attributes": {"weight": "100kg"}}],
            [{"data": b"", "attributes": {"weight": "100kg"}}],
            [{"attributes": {"weight": "100kg"}}],
        ],
    )
    def test_messages_validation_positive(self, messages):
        PubSubHook._validate_messages(messages)

    @pytest.mark.parametrize(
        "messages, error_message",
        [
            ([("wrong type",)], "Wrong message type. Must be a dictionary."),
            ([{"wrong_key": b"test"}], "Wrong message. Dictionary must contain 'data' or 'attributes'."),
            ([{"data": "wrong string"}], "Wrong message. 'data' must be send as a bytestring"),
            ([{"data": None}], "Wrong message. 'data' must be send as a bytestring"),
            (
                [{"attributes": None}],
                "Wrong message. If 'data' is not provided 'attributes' must be a non empty dictionary.",
            ),
            (
                [{"attributes": "wrong string"}],
                "Wrong message. If 'data' is not provided 'attributes' must be a non empty dictionary.",
            ),
        ],
    )
    def test_messages_validation_negative(self, messages, error_message):
        with pytest.raises(PubSubException) as ctx:
            PubSubHook._validate_messages(messages)
        assert str(ctx.value) == error_message


class TestPubSubAsyncHook:
    @pytest.fixture
    def hook(self):
        return PubSubAsyncHook()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook._get_subscriber_client")
    async def test_pull(self, mock_subscriber_client, hook):
        client = mock_subscriber_client.return_value

        await hook.pull(
            project_id=TEST_PROJECT, subscription=TEST_SUBSCRIPTION, max_messages=10, return_immediately=False
        )

        mock_subscriber_client.assert_called_once()
        client.pull.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                max_messages=10,
                return_immediately=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.pubsub.PubSubAsyncHook._get_subscriber_client")
    async def test_acknowledge(self, mock_subscriber_client, hook):
        client = mock_subscriber_client.return_value

        await hook.acknowledge(
            project_id=TEST_PROJECT,
            subscription=TEST_SUBSCRIPTION,
            messages=_generate_messages(3),
        )

        mock_subscriber_client.assert_called_once()
        client.acknowledge.assert_called_once_with(
            request=dict(
                subscription=EXPANDED_SUBSCRIPTION,
                ack_ids=["1", "2", "3"],
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.protobuf.field_mask_pb2 import FieldMask

from airflow.providers.google.cloud.operators.dataplex import DataplexHook

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
DATAPLEX_STRING = "airflow.providers.google.cloud.hooks.dataplex.{}"
DATAPLEX_HOOK_CLIENT = "airflow.providers.google.cloud.hooks.dataplex.DataplexHook.get_dataplex_client"
DATAPLEX_HOOK_DS_CLIENT = (
    "airflow.providers.google.cloud.hooks.dataplex.DataplexHook.get_dataplex_data_scan_client"
)
DATAPLEX_CATALOG_HOOK_CLIENT = (
    "airflow.providers.google.cloud.hooks.dataplex.DataplexHook.get_dataplex_catalog_client"
)

PROJECT_ID = "project-id"
REGION = "region"
LAKE_ID = "lake-id"
BODY = {"body": "test"}
DATAPLEX_TASK_ID = "testTask001"

GCP_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]

DATA_SCAN_ID = "test-data-scan-id"
ASSET_ID = "test_asset_id"
ZONE_ID = "test_zone_id"
JOB_ID = "job_id"

LOCATION = "us-central1"
ENTRY_ID = "entry-id"
ENTRY_GROUP_ID = "entry-group-id"
ENTRY_GROUP_BODY = {"description": "Some descr"}
ENTRY_GROUP_UPDATED_BODY = {"description": "Some new descr"}
ENTRY_TYPE_ID = "entry-type-id"
ENTRY_TYPE_BODY = {"description": "Some descr"}
ENTRY_TYPE_UPDATED_BODY = {"description": "Some new descr"}
ASPECT_TYPE_ID = "aspect-type-id"
ASPECT_TYPE_BODY = {"description": "Some descr"}
ASPECT_TYPE_UPDATED_BODY = {"description": "Some new descr"}
UPDATE_MASK = ["description"]
UPDATE_MASK_FOR_ENTRY = ["fully_qualified_name"]
ENTRY_ID_BODY = {
    "name": f"projects/{PROJECT_ID}/locations/{LOCATION}/entryGroups/{ENTRY_TYPE_ID}/entries/{ENTRY_ID}",
    "entry_type": f"projects/{PROJECT_ID}/locations/{LOCATION}/entryTypes/{ENTRY_TYPE_ID}",
}
ENTRY_ID_UPDATED_BODY = {
    "fully_qualified_name": f"dataplex:{PROJECT_ID}.{LOCATION}.{ENTRY_GROUP_ID}.some-entry"
}


COMMON_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}"
DATA_SCAN_NAME = f"projects/{PROJECT_ID}/locations/{REGION}/dataScans/{DATA_SCAN_ID}"
DATA_SCAN_JOB_NAME = f"projects/{PROJECT_ID}/locations/{REGION}/dataScans/{DATA_SCAN_ID}/jobs/{JOB_ID}"
ZONE_NAME = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}"
ZONE_PARENT = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}/zones/{ZONE_ID}"
ASSET_PARENT = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}/zones/{ZONE_ID}/assets/{ASSET_ID}"
DATASCAN_PARENT = f"projects/{PROJECT_ID}/locations/{REGION}"
ENTRY_GROUP_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}/entryGroup/{ENTRY_GROUP_ID}"
ENTRY_TYPE_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}/entryType/{ENTRY_TYPE_ID}"
ASPECT_TYPE_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}/aspectType/{ASPECT_TYPE_ID}"
ENTRY_PARENT = f"projects/{PROJECT_ID}/locations/{LOCATION}/entryGroups/{ENTRY_GROUP_ID}/entries/{ENTRY_ID}"


class TestDataplexHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"),
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = DataplexHook(
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_create_task(self, mock_client):
        self.hook.create_task(
            project_id=PROJECT_ID,
            region=REGION,
            lake_id=LAKE_ID,
            body=BODY,
            dataplex_task_id=DATAPLEX_TASK_ID,
            validate_only=None,
        )

        parent = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}"
        mock_client.return_value.create_task.assert_called_once_with(
            request=dict(
                parent=parent,
                task_id=DATAPLEX_TASK_ID,
                task=BODY,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_delete_task(self, mock_client):
        self.hook.delete_task(
            project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID, dataplex_task_id=DATAPLEX_TASK_ID
        )

        name = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}/tasks/{DATAPLEX_TASK_ID}"
        mock_client.return_value.delete_task.assert_called_once_with(
            request=dict(
                name=name,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_list_tasks(self, mock_client):
        self.hook.list_tasks(project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID)

        parent = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}"
        mock_client.return_value.list_tasks.assert_called_once_with(
            request=dict(
                parent=parent,
                page_size=None,
                page_token=None,
                filter=None,
                order_by=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_get_task(self, mock_client):
        self.hook.get_task(
            project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID, dataplex_task_id=DATAPLEX_TASK_ID
        )

        name = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}/tasks/{DATAPLEX_TASK_ID}"
        mock_client.return_value.get_task.assert_called_once_with(
            request=dict(
                name=name,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_create_lake(self, mock_client):
        self.hook.create_lake(
            project_id=PROJECT_ID,
            region=REGION,
            lake_id=LAKE_ID,
            body=BODY,
            validate_only=None,
        )

        parent = f"projects/{PROJECT_ID}/locations/{REGION}"
        mock_client.return_value.create_lake.assert_called_once_with(
            request=dict(
                parent=parent,
                lake_id=LAKE_ID,
                lake=BODY,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_delete_lake(self, mock_client):
        self.hook.delete_lake(project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID)

        name = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}"
        mock_client.return_value.delete_lake.assert_called_once_with(
            request=dict(
                name=name,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_get_lake(self, mock_client):
        self.hook.get_lake(project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID)

        name = f"projects/{PROJECT_ID}/locations/{REGION}/lakes/{LAKE_ID}/"
        mock_client.return_value.get_lake.assert_called_once_with(
            request=dict(
                name=name,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_create_zone(self, mock_client):
        self.hook.create_zone(project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID, zone_id=ZONE_ID, body={})

        mock_client.return_value.create_zone.assert_called_once_with(
            request=dict(
                parent=ZONE_NAME,
                zone_id=ZONE_ID,
                zone={},
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_delete_zone(self, mock_client):
        self.hook.delete_zone(project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID, zone_id=ZONE_ID)

        mock_client.return_value.delete_zone.assert_called_once_with(
            request=dict(
                name=ZONE_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_create_asset(self, mock_client):
        self.hook.create_asset(
            project_id=PROJECT_ID,
            region=REGION,
            lake_id=LAKE_ID,
            zone_id=ZONE_ID,
            asset_id=ASSET_ID,
            body={},
        )

        mock_client.return_value.create_asset.assert_called_once_with(
            request=dict(
                parent=ZONE_PARENT,
                asset={},
                asset_id=ASSET_ID,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_CLIENT)
    def test_delete_asset(self, mock_client):
        self.hook.delete_asset(
            project_id=PROJECT_ID, region=REGION, lake_id=LAKE_ID, zone_id=ZONE_ID, asset_id=ASSET_ID
        )

        mock_client.return_value.delete_asset.assert_called_once_with(
            request=dict(
                name=ASSET_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_DS_CLIENT)
    def test_create_data_scan(self, mock_client):
        self.hook.create_data_scan(project_id=PROJECT_ID, region=REGION, data_scan_id=DATA_SCAN_ID, body={})

        mock_client.return_value.create_data_scan.assert_called_once_with(
            request=dict(parent=DATASCAN_PARENT, data_scan_id=DATA_SCAN_ID, data_scan={}),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_DS_CLIENT)
    def test_run_data_scan(self, mock_client):
        self.hook.run_data_scan(project_id=PROJECT_ID, region=REGION, data_scan_id=DATA_SCAN_ID)

        mock_client.return_value.run_data_scan.assert_called_once_with(
            request=dict(
                name=DATA_SCAN_NAME,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_DS_CLIENT)
    def test_get_data_scan_job(self, mock_client):
        self.hook.get_data_scan_job(
            project_id=PROJECT_ID, region=REGION, job_id=JOB_ID, data_scan_id=DATA_SCAN_ID
        )

        mock_client.return_value.get_data_scan_job.assert_called_once_with(
            request=dict(name=DATA_SCAN_JOB_NAME, view="FULL"),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_DS_CLIENT)
    def test_delete_data_scan(self, mock_client):
        self.hook.delete_data_scan(project_id=PROJECT_ID, region=REGION, data_scan_id=DATA_SCAN_ID)

        mock_client.return_value.delete_data_scan.assert_called_once_with(
            request=dict(
                name=DATA_SCAN_NAME,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_HOOK_DS_CLIENT)
    def test_get_data_scan(self, mock_client):
        self.hook.get_data_scan(project_id=PROJECT_ID, region=REGION, data_scan_id=DATA_SCAN_ID)

        mock_client.return_value.get_data_scan.assert_called_once_with(
            request=dict(name=DATA_SCAN_NAME, view="FULL"),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_create_entry_group(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.create_entry_group(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_group_id=ENTRY_GROUP_ID,
            entry_group_configuration=ENTRY_GROUP_BODY,
            validate_only=False,
        )
        mock_client.return_value.create_entry_group.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                entry_group_id=ENTRY_GROUP_ID,
                entry_group=ENTRY_GROUP_BODY,
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_delete_entry_group(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_group_path
        mock_common_location_path.return_value = ENTRY_GROUP_PARENT
        self.hook.delete_entry_group(project_id=PROJECT_ID, location=LOCATION, entry_group_id=ENTRY_GROUP_ID)

        mock_client.return_value.delete_entry_group.assert_called_once_with(
            request=dict(
                name=ENTRY_GROUP_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_list_entry_groups(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.list_entry_groups(
            project_id=PROJECT_ID,
            location=LOCATION,
            order_by="name",
            page_size=2,
            filter_by="'description' = 'Some descr'",
        )
        mock_client.return_value.list_entry_groups.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                page_size=2,
                page_token=None,
                filter="'description' = 'Some descr'",
                order_by="name",
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_get_entry_group(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_group_path
        mock_common_location_path.return_value = ENTRY_GROUP_PARENT
        self.hook.get_entry_group(project_id=PROJECT_ID, location=LOCATION, entry_group_id=ENTRY_GROUP_ID)

        mock_client.return_value.get_entry_group.assert_called_once_with(
            request=dict(
                name=ENTRY_GROUP_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_update_entry_group(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_group_path
        mock_common_location_path.return_value = ENTRY_GROUP_PARENT
        self.hook.update_entry_group(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_group_id=ENTRY_GROUP_ID,
            entry_group_configuration=ENTRY_GROUP_UPDATED_BODY,
            update_mask=UPDATE_MASK,
            validate_only=False,
        )

        mock_client.return_value.update_entry_group.assert_called_once_with(
            request=dict(
                entry_group={**ENTRY_GROUP_UPDATED_BODY, "name": ENTRY_GROUP_PARENT},
                update_mask=FieldMask(paths=UPDATE_MASK),
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_create_entry_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.create_entry_type(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_type_id=ENTRY_TYPE_ID,
            entry_type_configuration=ENTRY_TYPE_BODY,
            validate_only=False,
        )
        mock_client.return_value.create_entry_type.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                entry_type_id=ENTRY_TYPE_ID,
                entry_type=ENTRY_TYPE_BODY,
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_delete_entry_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_type_path
        mock_common_location_path.return_value = ENTRY_TYPE_PARENT
        self.hook.delete_entry_type(project_id=PROJECT_ID, location=LOCATION, entry_type_id=ENTRY_TYPE_ID)

        mock_client.return_value.delete_entry_type.assert_called_once_with(
            request=dict(
                name=ENTRY_TYPE_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_list_entry_types(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.list_entry_types(
            project_id=PROJECT_ID,
            location=LOCATION,
            order_by="name",
            page_size=2,
            filter_by="'description' = 'Some descr'",
        )
        mock_client.return_value.list_entry_types.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                page_size=2,
                page_token=None,
                filter="'description' = 'Some descr'",
                order_by="name",
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_get_entry_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_type_path
        mock_common_location_path.return_value = ENTRY_TYPE_PARENT
        self.hook.get_entry_type(project_id=PROJECT_ID, location=LOCATION, entry_type_id=ENTRY_TYPE_ID)

        mock_client.return_value.get_entry_type.assert_called_once_with(
            request=dict(
                name=ENTRY_TYPE_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_update_entry_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_type_path
        mock_common_location_path.return_value = ENTRY_TYPE_PARENT
        self.hook.update_entry_type(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_type_id=ENTRY_TYPE_ID,
            entry_type_configuration=ENTRY_TYPE_UPDATED_BODY,
            update_mask=UPDATE_MASK,
            validate_only=False,
        )

        mock_client.return_value.update_entry_type.assert_called_once_with(
            request=dict(
                entry_type={**ENTRY_TYPE_UPDATED_BODY, "name": ENTRY_TYPE_PARENT},
                update_mask=FieldMask(paths=UPDATE_MASK),
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_create_aspect_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.create_aspect_type(
            project_id=PROJECT_ID,
            location=LOCATION,
            aspect_type_id=ASPECT_TYPE_ID,
            aspect_type_configuration=ASPECT_TYPE_BODY,
            validate_only=False,
        )
        mock_client.return_value.create_aspect_type.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                aspect_type_id=ASPECT_TYPE_ID,
                aspect_type=ASPECT_TYPE_BODY,
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_delete_aspect_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.aspect_type_path
        mock_common_location_path.return_value = ASPECT_TYPE_PARENT
        self.hook.delete_aspect_type(project_id=PROJECT_ID, location=LOCATION, aspect_type_id=ASPECT_TYPE_ID)

        mock_client.return_value.delete_aspect_type.assert_called_once_with(
            request=dict(
                name=ASPECT_TYPE_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_list_aspect_types(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.list_aspect_types(
            project_id=PROJECT_ID,
            location=LOCATION,
            order_by="name",
            page_size=2,
            filter_by="'description' = 'Some descr'",
        )
        mock_client.return_value.list_aspect_types.assert_called_once_with(
            request=dict(
                parent=COMMON_PARENT,
                page_size=2,
                page_token=None,
                filter="'description' = 'Some descr'",
                order_by="name",
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_get_aspect_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.aspect_type_path
        mock_common_location_path.return_value = ASPECT_TYPE_PARENT
        self.hook.get_aspect_type(project_id=PROJECT_ID, location=LOCATION, aspect_type_id=ASPECT_TYPE_ID)

        mock_client.return_value.get_aspect_type.assert_called_once_with(
            request=dict(
                name=ASPECT_TYPE_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_update_aspect_type(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_type_path
        mock_common_location_path.return_value = ENTRY_TYPE_PARENT
        self.hook.update_entry_type(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_type_id=ENTRY_TYPE_ID,
            entry_type_configuration=ENTRY_TYPE_UPDATED_BODY,
            update_mask=UPDATE_MASK,
            validate_only=False,
        )

        mock_client.return_value.update_entry_type.assert_called_once_with(
            request=dict(
                entry_type={**ENTRY_TYPE_UPDATED_BODY, "name": ENTRY_TYPE_PARENT},
                update_mask=FieldMask(paths=UPDATE_MASK),
                validate_only=False,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_create_entry(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_group_path
        mock_common_location_path.return_value = ENTRY_GROUP_PARENT
        self.hook.create_entry(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_id=ENTRY_ID,
            entry_group_id=ENTRY_GROUP_ID,
            entry_configuration=ENTRY_ID_BODY,
        )
        mock_client.return_value.create_entry.assert_called_once_with(
            request=dict(
                parent=ENTRY_GROUP_PARENT,
                entry_id=ENTRY_ID,
                entry=ENTRY_ID_BODY,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_delete_entry(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_path
        mock_common_location_path.return_value = ENTRY_PARENT
        self.hook.delete_entry(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_id=ENTRY_ID,
            entry_group_id=ENTRY_GROUP_ID,
        )

        mock_client.return_value.delete_entry.assert_called_once_with(
            request=dict(
                name=ENTRY_PARENT,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_list_entries(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_group_path
        mock_common_location_path.return_value = ENTRY_GROUP_PARENT
        self.hook.list_entries(
            project_id=PROJECT_ID,
            entry_group_id=ENTRY_GROUP_ID,
            location=LOCATION,
            page_size=2,
            filter_by="'description' = 'Some descr'",
        )
        mock_client.return_value.list_entries.assert_called_once_with(
            request=dict(
                parent=ENTRY_GROUP_PARENT,
                page_size=2,
                page_token=None,
                filter="'description' = 'Some descr'",
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_search_entries(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_common_location_path.return_value = COMMON_PARENT
        self.hook.search_entries(
            project_id=PROJECT_ID,
            location="global",
            query="displayname:Display Name",
        )
        mock_client.return_value.search_entries.assert_called_once_with(
            request=dict(
                name=COMMON_PARENT,
                query="displayname:Display Name",
                order_by=None,
                page_size=None,
                page_token=None,
                scope=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_get_entry(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_path
        mock_common_location_path.return_value = ENTRY_PARENT
        self.hook.get_entry(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_id=ENTRY_ID,
            entry_group_id=ENTRY_GROUP_ID,
        )

        mock_client.return_value.get_entry.assert_called_once_with(
            request=dict(
                name=ENTRY_PARENT,
                view=None,
                aspect_types=None,
                paths=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_lookup_entry(self, mock_client):
        mock_common_location_path = mock_client.return_value.common_location_path
        mock_entry_location_path = mock_client.return_value.entry_path
        mock_common_location_path.return_value = COMMON_PARENT
        mock_entry_location_path.return_value = ENTRY_PARENT
        self.hook.lookup_entry(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_id=ENTRY_ID,
            entry_group_id=ENTRY_GROUP_ID,
        )

        mock_client.return_value.lookup_entry.assert_called_once_with(
            request=dict(
                name=COMMON_PARENT,
                entry=ENTRY_PARENT,
                view=None,
                aspect_types=None,
                paths=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPLEX_CATALOG_HOOK_CLIENT)
    def test_update_entry(self, mock_client):
        mock_common_location_path = mock_client.return_value.entry_path
        mock_common_location_path.return_value = ENTRY_PARENT
        self.hook.update_entry(
            project_id=PROJECT_ID,
            location=LOCATION,
            entry_id=ENTRY_ID,
            entry_group_id=ENTRY_GROUP_ID,
            entry_configuration=ENTRY_ID_UPDATED_BODY,
            update_mask=UPDATE_MASK_FOR_ENTRY,
        )

        mock_client.return_value.update_entry.assert_called_once_with(
            request=dict(
                entry={**ENTRY_ID_UPDATED_BODY, "name": ENTRY_PARENT},
                update_mask=FieldMask(paths=UPDATE_MASK_FOR_ENTRY),
                allow_missing=False,
                delete_missing_aspects=False,
                aspect_keys=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT

from airflow.providers.google.cloud.hooks.dataproc_metastore import DataprocMetastoreHook

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_REGION: str = "test-region"
TEST_PROJECT_ID: str = "test-project-id"
TEST_BACKUP: str = "test-backup"
TEST_BACKUP_ID: str = "test-backup-id"
TEST_METADATA_IMPORT: dict = {
    "name": "test-metadata-import",
    "database_dump": {
        "gcs_uri": "gs://bucket_name/path_inside_bucket",
        "database_type": "MYSQL",
    },
}
TEST_METADATA_IMPORT_ID: str = "test-metadata-import-id"
TEST_SERVICE: dict = {"name": "test-service"}
TEST_SERVICE_ID: str = "test-service-id"
TEST_SERVICE_TO_UPDATE = {
    "labels": {
        "first_key": "first_value",
        "second_key": "second_value",
    }
}
TEST_UPDATE_MASK: dict = {"paths": ["labels"]}
TEST_PARENT: str = "projects/{}/locations/{}"
TEST_PARENT_SERVICES: str = "projects/{}/locations/{}/services/{}"
TEST_PARENT_BACKUPS: str = "projects/{}/locations/{}/services/{}/backups"
TEST_NAME_BACKUPS: str = "projects/{}/locations/{}/services/{}/backups/{}"
TEST_DESTINATION_GCS_FOLDER: str = "gs://bucket_name/path_inside_bucket"
TEST_TABLE_ID: str = "test_table"
TEST_PARTITION_NAME = "column=value"
TEST_SUBPARTITION_NAME = "column1=value1/column2=value2"
TEST_PARTITIONS_QUERY_ALL = """
    SELECT PARTITIONS.*, TBLS.TBL_TYPE, TBLS.TBL_NAME
    FROM PARTITIONS
    INNER JOIN TBLS ON PARTITIONS.TBL_ID = TBLS.TBL_ID
    WHERE TBLS.TBL_NAME = '{}';"""

TEST_PARTITIONS_QUERY = """
    SELECT PARTITIONS.*, TBLS.TBL_TYPE, TBLS.TBL_NAME
    FROM PARTITIONS
    INNER JOIN TBLS ON PARTITIONS.TBL_ID = TBLS.TBL_ID
    WHERE TBLS.TBL_NAME = '{}'
        AND PARTITIONS.PART_NAME IN ({});"""
BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
DATAPROC_METASTORE_STRING = "airflow.providers.google.cloud.hooks.dataproc_metastore.{}"


class TestDataprocMetastoreWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = DataprocMetastoreHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_backup(self, mock_client) -> None:
        self.hook.create_backup(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup=TEST_BACKUP,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_backup.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                backup=TEST_BACKUP,
                backup_id=TEST_BACKUP_ID,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_metadata_import(self, mock_client) -> None:
        self.hook.create_metadata_import(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            metadata_import=TEST_METADATA_IMPORT,
            metadata_import_id=TEST_METADATA_IMPORT_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_metadata_import.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                metadata_import=TEST_METADATA_IMPORT,
                metadata_import_id=TEST_METADATA_IMPORT_ID,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_service(self, mock_client) -> None:
        self.hook.create_service(
            region=TEST_REGION,
            project_id=TEST_PROJECT_ID,
            service=TEST_SERVICE,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_service.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT.format(TEST_PROJECT_ID, TEST_REGION),
                service_id=TEST_SERVICE_ID,
                service=TEST_SERVICE,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_get_backup(self, mock_client) -> None:
        self.hook.get_backup(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.get_backup.assert_called_once_with(
            request=dict(
                name=TEST_NAME_BACKUPS.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID, TEST_BACKUP_ID),
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_delete_backup(self, mock_client) -> None:
        self.hook.delete_backup(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_backup.assert_called_once_with(
            request=dict(
                name=TEST_NAME_BACKUPS.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID, TEST_BACKUP_ID),
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_delete_service(self, mock_client) -> None:
        self.hook.delete_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_service.assert_called_once_with(
            request=dict(
                name=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                request_id=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_export_metadata(self, mock_client) -> None:
        self.hook.export_metadata(
            destination_gcs_folder=TEST_DESTINATION_GCS_FOLDER,
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.export_metadata.assert_called_once_with(
            request=dict(
                destination_gcs_folder=TEST_DESTINATION_GCS_FOLDER,
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                request_id=None,
                database_dump_type=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_get_service(self, mock_client) -> None:
        self.hook.get_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.get_service.assert_called_once_with(
            request=dict(
                name=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_list_backups(self, mock_client) -> None:
        self.hook.list_backups(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.list_backups.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_BACKUPS.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                page_size=None,
                page_token=None,
                filter=None,
                order_by=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_restore_service(self, mock_client) -> None:
        self.hook.restore_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup_project_id=TEST_PROJECT_ID,
            backup_region=TEST_REGION,
            backup_service_id=TEST_SERVICE_ID,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.restore_service.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                backup=TEST_NAME_BACKUPS.format(
                    TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID, TEST_BACKUP_ID
                ),
                restore_type=None,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_update_service(self, mock_client) -> None:
        self.hook.update_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            service=TEST_SERVICE_TO_UPDATE,
            update_mask=TEST_UPDATE_MASK,
        )
        mock_client.assert_called_once()
        mock_client.return_value.update_service.assert_called_once_with(
            request=dict(
                service=TEST_SERVICE_TO_UPDATE,
                update_mask=TEST_UPDATE_MASK,
                request_id=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @pytest.mark.parametrize(
        "partitions_input, partitions",
        [
            ([TEST_PARTITION_NAME], f"'{TEST_PARTITION_NAME}'"),
            ([TEST_SUBPARTITION_NAME], f"'{TEST_SUBPARTITION_NAME}'"),
            (
                [TEST_PARTITION_NAME, TEST_SUBPARTITION_NAME],
                f"'{TEST_PARTITION_NAME}', '{TEST_SUBPARTITION_NAME}'",
            ),
            ([TEST_PARTITION_NAME, TEST_PARTITION_NAME], f"'{TEST_PARTITION_NAME}'"),
        ],
    )
    @mock.patch(
        DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client_v1beta")
    )
    def test_list_hive_partitions(self, mock_client, partitions_input, partitions) -> None:
        self.hook.list_hive_partitions(
            project_id=TEST_PROJECT_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE_ID,
            partition_names=partitions_input,
        )
        mock_client.assert_called_once()
        mock_client.return_value.query_metadata.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                query=TEST_PARTITIONS_QUERY.format(TEST_TABLE_ID, partitions),
            ),
        )

    @pytest.mark.parametrize("partitions", [[], None])
    @mock.patch(
        DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client_v1beta")
    )
    def test_list_hive_partitions_empty_list(self, mock_client, partitions) -> None:
        self.hook.list_hive_partitions(
            project_id=TEST_PROJECT_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE_ID,
            partition_names=partitions,
        )
        mock_client.assert_called_once()
        mock_client.return_value.query_metadata.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                query=TEST_PARTITIONS_QUERY_ALL.format(TEST_TABLE_ID),
            ),
        )


class TestDataprocMetastoreWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = DataprocMetastoreHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_backup(self, mock_client) -> None:
        self.hook.create_backup(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup=TEST_BACKUP,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_backup.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                backup=TEST_BACKUP,
                backup_id=TEST_BACKUP_ID,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_metadata_import(self, mock_client) -> None:
        self.hook.create_metadata_import(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            metadata_import=TEST_METADATA_IMPORT,
            metadata_import_id=TEST_METADATA_IMPORT_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_metadata_import.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                metadata_import=TEST_METADATA_IMPORT,
                metadata_import_id=TEST_METADATA_IMPORT_ID,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_create_service(self, mock_client) -> None:
        self.hook.create_service(
            region=TEST_REGION,
            project_id=TEST_PROJECT_ID,
            service=TEST_SERVICE,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.create_service.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT.format(TEST_PROJECT_ID, TEST_REGION),
                service_id=TEST_SERVICE_ID,
                service=TEST_SERVICE,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_delete_backup(self, mock_client) -> None:
        self.hook.delete_backup(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_backup.assert_called_once_with(
            request=dict(
                name=TEST_NAME_BACKUPS.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID, TEST_BACKUP_ID),
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_delete_service(self, mock_client) -> None:
        self.hook.delete_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.delete_service.assert_called_once_with(
            request=dict(
                name=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                request_id=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_export_metadata(self, mock_client) -> None:
        self.hook.export_metadata(
            destination_gcs_folder=TEST_DESTINATION_GCS_FOLDER,
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.export_metadata.assert_called_once_with(
            request=dict(
                destination_gcs_folder=TEST_DESTINATION_GCS_FOLDER,
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                request_id=None,
                database_dump_type=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_get_service(self, mock_client) -> None:
        self.hook.get_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.get_service.assert_called_once_with(
            request=dict(
                name=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_list_backups(self, mock_client) -> None:
        self.hook.list_backups(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.list_backups.assert_called_once_with(
            request=dict(
                parent=TEST_PARENT_BACKUPS.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                page_size=None,
                page_token=None,
                filter=None,
                order_by=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_restore_service(self, mock_client) -> None:
        self.hook.restore_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            backup_project_id=TEST_PROJECT_ID,
            backup_region=TEST_REGION,
            backup_service_id=TEST_SERVICE_ID,
            backup_id=TEST_BACKUP_ID,
        )
        mock_client.assert_called_once()
        mock_client.return_value.restore_service.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                backup=TEST_NAME_BACKUPS.format(
                    TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID, TEST_BACKUP_ID
                ),
                restore_type=None,
                request_id=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client"))
    def test_update_service(self, mock_client) -> None:
        self.hook.update_service(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            service_id=TEST_SERVICE_ID,
            service=TEST_SERVICE_TO_UPDATE,
            update_mask=TEST_UPDATE_MASK,
        )
        mock_client.assert_called_once()
        mock_client.return_value.update_service.assert_called_once_with(
            request=dict(
                service=TEST_SERVICE_TO_UPDATE,
                update_mask=TEST_UPDATE_MASK,
                request_id=None,
            ),
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @pytest.mark.parametrize(
        "partitions_input, partitions",
        [
            ([TEST_PARTITION_NAME], f"'{TEST_PARTITION_NAME}'"),
            ([TEST_SUBPARTITION_NAME], f"'{TEST_SUBPARTITION_NAME}'"),
            (
                [TEST_PARTITION_NAME, TEST_SUBPARTITION_NAME],
                f"'{TEST_PARTITION_NAME}', '{TEST_SUBPARTITION_NAME}'",
            ),
            ([TEST_PARTITION_NAME, TEST_PARTITION_NAME], f"'{TEST_PARTITION_NAME}'"),
        ],
    )
    @mock.patch(
        DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client_v1beta")
    )
    def test_list_hive_partitions(self, mock_client, partitions_input, partitions) -> None:
        self.hook.list_hive_partitions(
            project_id=TEST_PROJECT_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE_ID,
            partition_names=partitions_input,
        )
        mock_client.assert_called_once()
        mock_client.return_value.query_metadata.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                query=TEST_PARTITIONS_QUERY.format(TEST_TABLE_ID, partitions),
            ),
        )

    @pytest.mark.parametrize("partitions", [[], None])
    @mock.patch(
        DATAPROC_METASTORE_STRING.format("DataprocMetastoreHook.get_dataproc_metastore_client_v1beta")
    )
    def test_list_hive_partitions_empty_list(self, mock_client, partitions) -> None:
        self.hook.list_hive_partitions(
            project_id=TEST_PROJECT_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE_ID,
            partition_names=partitions,
        )
        mock_client.assert_called_once()
        mock_client.return_value.query_metadata.assert_called_once_with(
            request=dict(
                service=TEST_PARENT_SERVICES.format(TEST_PROJECT_ID, TEST_REGION, TEST_SERVICE_ID),
                query=TEST_PARTITIONS_QUERY_ALL.format(TEST_TABLE_ID),
            ),
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
import datetime
import logging
from asyncio import Future
from unittest import mock

import pytest
from google.cloud.container_v1.types import Operation
from kubernetes.client import models as k8s

try:
    from airflow.providers.cncf.kubernetes.triggers.pod import ContainerState
except ImportError:
    # preserve backward compatibility for older versions of cncf.kubernetes provider, remove this when minimum cncf.kubernetes provider is 10.0
    from airflow.providers.cncf.kubernetes.triggers.kubernetes_pod import (  # type: ignore[no-redef]
        ContainerState,
    )
from airflow.providers.google.cloud.triggers.kubernetes_engine import (
    GKEJobTrigger,
    GKEOperationTrigger,
    GKEStartPodTrigger,
)
from airflow.triggers.base import TriggerEvent

GKE_TRIGGERS_PATH = "airflow.providers.google.cloud.triggers.kubernetes_engine"
TRIGGER_GKE_POD_PATH = GKE_TRIGGERS_PATH + ".GKEStartPodTrigger"
TRIGGER_GKE_JOB_PATH = GKE_TRIGGERS_PATH + ".GKEJobTrigger"
TRIGGER_KUB_POD_PATH = "airflow.providers.cncf.kubernetes.triggers.pod.KubernetesPodTrigger"
POD_NAME = "test-pod-name"
JOB_NAME = "test-job-name"
NAMESPACE = "default"
POLL_INTERVAL = 2
CLUSTER_CONTEXT = "test-context"
IN_CLUSTER = False
SHOULD_DELETE_POD = True
GET_LOGS = True
STARTUP_TIMEOUT_SECS = 120
TRIGGER_START_TIME = datetime.datetime.now(tz=datetime.timezone.utc)
CLUSTER_URL = "https://test-host"
SSL_CA_CERT = "TEST_SSL_CA_CERT_CONTENT"
FAILED_RESULT_MSG = "Test message that appears when trigger have failed event."
BASE_CONTAINER_NAME = "base"
ON_FINISH_ACTION = "delete_pod"
XCOM_PUSH = False

OPERATION_NAME = "test-operation-name"
PROJECT_ID = "test-project-id"
LOCATION = "us-central1-c"
GCP_CONN_ID = "test-non-existing-project-id"
IMPERSONATION_CHAIN = ["impersonate", "this", "test"]
TRIGGER_PATH = f"{GKE_TRIGGERS_PATH}.GKEOperationTrigger"
EXC_MSG = "test error msg"


@pytest.fixture
def trigger():
    return GKEStartPodTrigger(
        pod_name=POD_NAME,
        pod_namespace=NAMESPACE,
        poll_interval=POLL_INTERVAL,
        cluster_context=CLUSTER_CONTEXT,
        in_cluster=IN_CLUSTER,
        get_logs=GET_LOGS,
        startup_timeout=STARTUP_TIMEOUT_SECS,
        trigger_start_time=TRIGGER_START_TIME,
        cluster_url=CLUSTER_URL,
        ssl_ca_cert=SSL_CA_CERT,
        base_container_name=BASE_CONTAINER_NAME,
        gcp_conn_id=GCP_CONN_ID,
        impersonation_chain=IMPERSONATION_CHAIN,
        on_finish_action=ON_FINISH_ACTION,
    )


@pytest.fixture
def job_trigger():
    return GKEJobTrigger(
        cluster_url=CLUSTER_URL,
        ssl_ca_cert=SSL_CA_CERT,
        job_name=JOB_NAME,
        job_namespace=NAMESPACE,
        pod_names=[
            POD_NAME,
        ],
        pod_namespace=NAMESPACE,
        base_container_name=BASE_CONTAINER_NAME,
        gcp_conn_id=GCP_CONN_ID,
        poll_interval=POLL_INTERVAL,
        impersonation_chain=IMPERSONATION_CHAIN,
        get_logs=GET_LOGS,
        do_xcom_push=XCOM_PUSH,
    )


class TestGKEStartPodTrigger:
    @staticmethod
    def _mock_pod_result(result_to_mock):
        f = Future()
        f.set_result(result_to_mock)
        return f

    def test_serialize_should_execute_successfully(self, trigger):
        classpath, kwargs_dict = trigger.serialize()

        assert classpath == TRIGGER_GKE_POD_PATH
        assert kwargs_dict == {
            "pod_name": POD_NAME,
            "pod_namespace": NAMESPACE,
            "poll_interval": POLL_INTERVAL,
            "cluster_context": CLUSTER_CONTEXT,
            "in_cluster": IN_CLUSTER,
            "get_logs": GET_LOGS,
            "startup_timeout": STARTUP_TIMEOUT_SECS,
            "trigger_start_time": TRIGGER_START_TIME,
            "cluster_url": CLUSTER_URL,
            "ssl_ca_cert": SSL_CA_CERT,
            "base_container_name": BASE_CONTAINER_NAME,
            "on_finish_action": ON_FINISH_ACTION,
            "should_delete_pod": SHOULD_DELETE_POD,
            "gcp_conn_id": GCP_CONN_ID,
            "impersonation_chain": IMPERSONATION_CHAIN,
            "last_log_time": None,
            "logging_interval": None,
        }

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}._wait_for_pod_start")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_run_loop_return_success_event_should_execute_successfully(
        self, mock_hook, mock_wait_pod, trigger
    ):
        mock_hook.get_pod.return_value = self._mock_pod_result(mock.MagicMock())
        mock_wait_pod.return_value = ContainerState.TERMINATED

        expected_event = TriggerEvent(
            {
                "name": POD_NAME,
                "namespace": NAMESPACE,
                "status": "success",
                "message": "All containers inside pod have started successfully.",
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}._wait_for_pod_start")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_run_loop_return_failed_event_should_execute_successfully(
        self, mock_hook, mock_wait_pod, trigger
    ):
        mock_hook.get_pod.return_value = self._mock_pod_result(
            mock.MagicMock(
                status=mock.MagicMock(
                    message=FAILED_RESULT_MSG,
                )
            )
        )
        mock_wait_pod.return_value = ContainerState.FAILED

        expected_event = TriggerEvent(
            {
                "name": POD_NAME,
                "namespace": NAMESPACE,
                "status": "failed",
                "message": "pod failed",
            }
        )
        actual_event = await trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}._wait_for_pod_start")
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}.define_container_state")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_run_loop_return_waiting_event_should_execute_successfully(
        self, mock_hook, mock_method, mock_wait_pod, trigger, caplog
    ):
        mock_hook.get_pod.return_value = self._mock_pod_result(mock.MagicMock())
        mock_method.return_value = ContainerState.WAITING

        caplog.set_level(logging.INFO)

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert not task.done()
        assert "Container is not completed and still working."
        assert f"Sleeping for {POLL_INTERVAL} seconds."

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}._wait_for_pod_start")
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}.define_container_state")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_run_loop_return_running_event_should_execute_successfully(
        self, mock_hook, mock_method, mock_wait_pod, trigger, caplog
    ):
        mock_hook.get_pod.return_value = self._mock_pod_result(mock.MagicMock())
        mock_method.return_value = ContainerState.RUNNING

        caplog.set_level(logging.INFO)

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert not task.done()
        assert "Container is not completed and still working."
        assert f"Sleeping for {POLL_INTERVAL} seconds."

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}._wait_for_pod_start")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_logging_in_trigger_when_exception_should_execute_successfully(
        self, mock_hook, mock_wait_pod, trigger, caplog
    ):
        """
        Test that GKEStartPodTrigger fires the correct event in case of an error.
        """
        mock_hook.get_pod.side_effect = Exception("Test exception")

        generator = trigger.run()
        actual = await generator.asend(None)
        actual_stack_trace = actual.payload.pop("stack_trace")
        assert (
            TriggerEvent(
                {"name": POD_NAME, "namespace": NAMESPACE, "status": "error", "message": "Test exception"}
            )
            == actual
        )
        assert actual_stack_trace.startswith("Traceback (most recent call last):")

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_KUB_POD_PATH}.define_container_state")
    @mock.patch(f"{TRIGGER_GKE_POD_PATH}.hook")
    async def test_logging_in_trigger_when_fail_should_execute_successfully(
        self, mock_hook, mock_method, trigger, caplog
    ):
        """
        Test that GKEStartPodTrigger fires the correct event in case of fail.
        """
        mock_hook.get_pod.return_value = self._mock_pod_result(mock.MagicMock())
        mock_method.return_value = ContainerState.FAILED
        caplog.set_level(logging.INFO)

        generator = trigger.run()
        await generator.asend(None)
        assert "Container logs:"

    @pytest.mark.parametrize(
        "container_state, expected_state",
        [
            (
                {"running": k8s.V1ContainerStateRunning(), "terminated": None, "waiting": None},
                ContainerState.RUNNING,
            ),
            (
                {"running": None, "terminated": k8s.V1ContainerStateTerminated(exit_code=0), "waiting": None},
                ContainerState.TERMINATED,
            ),
            (
                {"running": None, "terminated": None, "waiting": k8s.V1ContainerStateWaiting()},
                ContainerState.WAITING,
            ),
        ],
    )
    def test_define_container_state_should_execute_successfully(
        self, trigger, container_state, expected_state
    ):
        pod = k8s.V1Pod(
            metadata=k8s.V1ObjectMeta(name="base", namespace="default"),
            status=k8s.V1PodStatus(
                container_statuses=[
                    k8s.V1ContainerStatus(
                        name="base",
                        image="alpine",
                        image_id="1",
                        ready=True,
                        restart_count=1,
                        state=k8s.V1ContainerState(**container_state),
                    )
                ]
            ),
        )

        assert expected_state == trigger.define_container_state(pod)


@pytest.fixture
def operation_trigger():
    return GKEOperationTrigger(
        operation_name=OPERATION_NAME,
        project_id=PROJECT_ID,
        location=LOCATION,
        gcp_conn_id=GCP_CONN_ID,
        impersonation_chain=IMPERSONATION_CHAIN,
        poll_interval=POLL_INTERVAL,
    )


@pytest.fixture
def async_get_operation_result():
    def func(**kwargs):
        m = mock.MagicMock()
        m.configure_mock(**kwargs)
        f = Future()
        f.set_result(m)
        return f

    return func


class TestGKEOperationTrigger:
    def test_serialize(self, operation_trigger):
        classpath, trigger_init_kwargs = operation_trigger.serialize()
        assert classpath == TRIGGER_PATH
        assert trigger_init_kwargs == {
            "operation_name": OPERATION_NAME,
            "project_id": PROJECT_ID,
            "location": LOCATION,
            "gcp_conn_id": GCP_CONN_ID,
            "impersonation_chain": IMPERSONATION_CHAIN,
            "poll_interval": POLL_INTERVAL,
        }

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_success_event(
        self, mock_hook, operation_trigger, async_get_operation_result
    ):
        mock_hook.return_value.get_operation.return_value = async_get_operation_result(
            name=OPERATION_NAME,
            status=Operation.Status.DONE,
        )

        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": "Operation is successfully ended.",
                "operation_name": OPERATION_NAME,
            }
        )
        actual_event = await operation_trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_failed_event_status_unspecified(
        self,
        mock_hook,
        operation_trigger,
        async_get_operation_result,
    ):
        mock_hook.return_value.get_operation.return_value = async_get_operation_result(
            name=OPERATION_NAME,
            status=Operation.Status.STATUS_UNSPECIFIED,
        )

        expected_event = TriggerEvent(
            {
                "status": "failed",
                "message": f"Operation has failed with status: {Operation.Status.STATUS_UNSPECIFIED}",
            }
        )
        actual_event = await operation_trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_failed_event_status_aborting(
        self,
        mock_hook,
        operation_trigger,
        async_get_operation_result,
    ):
        mock_hook.return_value.get_operation.return_value = async_get_operation_result(
            name=OPERATION_NAME,
            status=Operation.Status.ABORTING,
        )

        expected_event = TriggerEvent(
            {
                "status": "failed",
                "message": f"Operation has failed with status: {Operation.Status.ABORTING}",
            }
        )
        actual_event = await operation_trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_error_event(
        self, mock_hook, operation_trigger, async_get_operation_result
    ):
        mock_hook.return_value.get_operation.side_effect = Exception(EXC_MSG)

        expected_event = TriggerEvent(
            {
                "status": "error",
                "message": EXC_MSG,
            }
        )
        actual_event = await operation_trigger.run().asend(None)

        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_waiting_event_pending_status(
        self,
        mock_hook,
        operation_trigger,
        async_get_operation_result,
        caplog,
    ):
        mock_hook.return_value.get_operation.return_value = async_get_operation_result(
            name=OPERATION_NAME,
            status=Operation.Status.PENDING,
        )

        caplog.set_level(logging.INFO)

        task = asyncio.create_task(operation_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert not task.done()
        assert "Operation is still running."
        assert f"Sleeping for {POLL_INTERVAL}s..."

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_PATH}._get_hook")
    async def test_run_loop_return_waiting_event_running_status(
        self,
        mock_hook,
        operation_trigger,
        async_get_operation_result,
        caplog,
    ):
        mock_hook.return_value.get_operation.return_value = async_get_operation_result(
            name=OPERATION_NAME,
            status=Operation.Status.RUNNING,
        )

        caplog.set_level(logging.INFO)

        task = asyncio.create_task(operation_trigger.run().__anext__())
        await asyncio.sleep(0.5)

        assert not task.done()
        assert "Operation is still running."
        assert f"Sleeping for {POLL_INTERVAL}s..."


class TestGKEStartJobTrigger:
    def test_serialize(self, job_trigger):
        classpath, kwargs_dict = job_trigger.serialize()

        assert classpath == TRIGGER_GKE_JOB_PATH
        assert kwargs_dict == {
            "cluster_url": CLUSTER_URL,
            "ssl_ca_cert": SSL_CA_CERT,
            "job_name": JOB_NAME,
            "job_namespace": NAMESPACE,
            "pod_names": [
                POD_NAME,
            ],
            "pod_namespace": NAMESPACE,
            "base_container_name": BASE_CONTAINER_NAME,
            "gcp_conn_id": GCP_CONN_ID,
            "poll_interval": POLL_INTERVAL,
            "impersonation_chain": IMPERSONATION_CHAIN,
            "get_logs": GET_LOGS,
            "do_xcom_push": XCOM_PUSH,
        }

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_GKE_JOB_PATH}.hook")
    async def test_run_success(self, mock_hook, job_trigger):
        mock_job = mock.MagicMock()
        mock_job.metadata.name = JOB_NAME
        mock_job.metadata.namespace = NAMESPACE
        mock_hook.wait_until_job_complete.side_effect = mock.AsyncMock(return_value=mock_job)

        mock_pod = mock.MagicMock()
        mock_pod.metadata.name = POD_NAME
        mock_pod.metadata.namespace = NAMESPACE
        mock_hook.get_pod.side_effect = mock.AsyncMock(return_value=mock_pod)

        mock_is_job_failed = mock_hook.is_job_failed
        mock_is_job_failed.return_value = False

        mock_job_dict = mock_job.to_dict.return_value

        event_actual = await job_trigger.run().asend(None)

        mock_hook.wait_until_job_complete.assert_called_once_with(
            name=JOB_NAME, namespace=NAMESPACE, poll_interval=POLL_INTERVAL
        )
        mock_job.to_dict.assert_called_once()
        mock_is_job_failed.assert_called_once_with(job=mock_job)
        assert event_actual == TriggerEvent(
            {
                "name": JOB_NAME,
                "namespace": NAMESPACE,
                "pod_names": [
                    POD_NAME,
                ],
                "pod_namespace": NAMESPACE,
                "status": "success",
                "message": "Job completed successfully",
                "job": mock_job_dict,
                "xcom_result": None,
            }
        )

    @pytest.mark.asyncio
    @mock.patch(f"{TRIGGER_GKE_JOB_PATH}.hook")
    async def test_run_fail(self, mock_hook, job_trigger):
        mock_job = mock.MagicMock()
        mock_job.metadata.name = JOB_NAME
        mock_job.metadata.namespace = NAMESPACE
        mock_hook.wait_until_job_complete.side_effect = mock.AsyncMock(return_value=mock_job)

        mock_pod = mock.MagicMock()
        mock_pod.metadata.name = POD_NAME
        mock_pod.metadata.namespace = NAMESPACE
        mock_hook.get_pod.side_effect = mock.AsyncMock(return_value=mock_pod)

        mock_is_job_failed = mock_hook.is_job_failed
        mock_is_job_failed.return_value = "Error"

        mock_job_dict = mock_job.to_dict.return_value

        event_actual = await job_trigger.run().asend(None)

        mock_hook.wait_until_job_complete.assert_called_once_with(
            name=JOB_NAME, namespace=NAMESPACE, poll_interval=POLL_INTERVAL
        )
        mock_job.to_dict.assert_called_once()
        mock_is_job_failed.assert_called_once_with(job=mock_job)
        assert event_actual == TriggerEvent(
            {
                "name": JOB_NAME,
                "namespace": NAMESPACE,
                "pod_names": [
                    POD_NAME,
                ],
                "pod_namespace": NAMESPACE,
                "status": "error",
                "message": "Job failed with error: Error",
                "job": mock_job_dict,
                "xcom_result": None,
            }
        )

    @mock.patch(f"{GKE_TRIGGERS_PATH}.GKEKubernetesAsyncHook")
    def test_hook(self, mock_hook, job_trigger):
        hook_expected = mock_hook.return_value

        hook_actual = job_trigger.hook

        mock_hook.assert_called_once_with(
            cluster_url=CLUSTER_URL,
            ssl_ca_cert=SSL_CA_CERT,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        assert hook_actual == hook_expected



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import urllib.parse

import pytest

from airflow.providers.common.compat.assets import Asset
from airflow.providers.google.assets.bigquery import convert_asset_to_openlineage, create_asset, sanitize_uri


def test_sanitize_uri_pass() -> None:
    uri_i = urllib.parse.urlsplit("bigquery://project/dataset/table")
    uri_o = sanitize_uri(uri_i)
    assert urllib.parse.urlunsplit(uri_o) == "bigquery://project/dataset/table"


@pytest.mark.parametrize(
    "value",
    [
        pytest.param("bigquery://", id="blank"),
        pytest.param("bigquery:///dataset/table", id="no-project"),
        pytest.param("bigquery://project/dataset", id="missing-component"),
        pytest.param("bigquery://project/dataset/table/column", id="extra-component"),
    ],
)
def test_sanitize_uri_fail(value: str) -> None:
    uri_i = urllib.parse.urlsplit(value)
    with pytest.raises(ValueError):
        sanitize_uri(uri_i)


def test_create_asset():
    assert create_asset(project_id="project", dataset_id="dataset", table_id="table") == Asset(
        uri="bigquery://project/dataset/table"
    )


def test_convert_asset_to_openlineage():
    uri = "bigquery://project/dataset/table"
    ol_dataset = convert_asset_to_openlineage(asset=Asset(uri=uri), lineage_context=None)
    assert ol_dataset.namespace == "bigquery"
    assert ol_dataset.name == "project.dataset.table"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains operator to move data from HTTP endpoint to GCS."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator
from airflow.providers.http.hooks.http import HttpHook

if TYPE_CHECKING:
    from collections.abc import Sequence

    from requests.auth import AuthBase

    from airflow.utils.context import Context


class HttpToGCSOperator(BaseOperator):
    """
    Calls an endpoint on an HTTP system to execute an action and store the result in GCS.

    :param http_conn_id: The :ref:`http connection<howto/connection:http>` to run
        the operator against
    :param endpoint: The relative part of the full url. (templated)
    :param method: The HTTP method to use, default = "POST"
    :param data: The data to pass. POST-data in POST/PUT and params
        in the URL for a GET request. (templated)
    :param headers: The HTTP headers to be added to the GET request
    :param response_check: A check against the 'requests' response object.
        The callable takes the response object as the first positional argument
        and optionally any number of keyword arguments available in the context dictionary.
        It should return True for 'pass' and False otherwise.
    :param response_filter: A function allowing you to manipulate the response
        text. e.g response_filter=lambda response: json.loads(response.text).
        The callable takes the response object as the first positional argument
        and optionally any number of keyword arguments available in the context dictionary.
    :param extra_options: Extra options for the 'requests' library, see the
        'requests' documentation (options to modify timeout, ssl, etc.)
    :param log_response: Log the response (default: False)
    :param auth_type: The auth type for the service
    :param tcp_keep_alive: Enable TCP Keep Alive for the connection.
    :param tcp_keep_alive_idle: The TCP Keep Alive Idle parameter (corresponds to ``socket.TCP_KEEPIDLE``).
    :param tcp_keep_alive_count: The TCP Keep Alive count parameter (corresponds to ``socket.TCP_KEEPCNT``)
    :param tcp_keep_alive_interval: The TCP Keep Alive interval parameter (corresponds to
        ``socket.TCP_KEEPINTVL``)
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term credentials,
        or chained list of accounts required to get the access_token of the last account in the list,
        which will be impersonated in the request. If set as a string,
        the account must grant the originating account the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant Service Account Token Creator IAM role to the directly preceding identity,
        with first account from the list granting this role to the originating account.
    :param bucket_name: The bucket to upload to.
    :param object_name: The object name to set when uploading the file.
    :param mime_type: The file mime type set when uploading the file.
    :param gzip: Option to compress local file or file data for upload
    :param encoding: bytes encoding for file data if provided as string
    :param chunk_size: Blob chunk size.
    :param timeout: Request timeout in seconds.
    :param num_max_attempts: Number of attempts to try to upload the file.
    :param metadata: The metadata to be uploaded with the file.
    :param cache_contro: Cache-Control metadata field.
    :param user_project: The identifier of the Google Cloud project to bill for the request. Required for Requester Pays buckets.
    """

    template_fields: Sequence[str] = (
        "http_conn_id",
        "endpoint",
        "data",
        "headers",
        "gcp_conn_id",
        "bucket_name",
        "object_name",
    )
    template_fields_renderers = {"headers": "json", "data": "py"}
    template_ext: Sequence[str] = ()
    ui_color = "#f4a460"

    def __init__(
        self,
        *,
        endpoint: str | None = None,
        method: str = "GET",
        data: Any = None,
        headers: dict[str, str] | None = None,
        extra_options: dict[str, Any] | None = None,
        http_conn_id: str = "http_default",
        log_response: bool = False,
        auth_type: type[AuthBase] | None = None,
        tcp_keep_alive: bool = True,
        tcp_keep_alive_idle: int = 120,
        tcp_keep_alive_count: int = 20,
        tcp_keep_alive_interval: int = 30,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        bucket_name: str,
        object_name: str,
        mime_type: str | None = None,
        gzip: bool = False,
        encoding: str | None = None,
        chunk_size: int | None = None,
        timeout: int | None = None,
        num_max_attempts: int = 3,
        metadata: dict | None = None,
        cache_control: str | None = None,
        user_project: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.http_conn_id = http_conn_id
        self.method = method
        self.endpoint = endpoint
        self.headers = headers or {}
        self.data = data or {}
        self.extra_options = extra_options or {}
        self.log_response = log_response
        self.auth_type = auth_type
        self.tcp_keep_alive = tcp_keep_alive
        self.tcp_keep_alive_idle = tcp_keep_alive_idle
        self.tcp_keep_alive_count = tcp_keep_alive_count
        self.tcp_keep_alive_interval = tcp_keep_alive_interval
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.mime_type = mime_type
        self.gzip = gzip
        self.encoding = encoding
        self.chunk_size = chunk_size
        self.timeout = timeout
        self.num_max_attempts = num_max_attempts
        self.metadata = metadata
        self.cache_control = cache_control
        self.user_project = user_project

    @cached_property
    def http_hook(self) -> HttpHook:
        """Create and return an HttpHook."""
        return HttpHook(
            self.method,
            http_conn_id=self.http_conn_id,
            auth_type=self.auth_type,
            tcp_keep_alive=self.tcp_keep_alive,
            tcp_keep_alive_idle=self.tcp_keep_alive_idle,
            tcp_keep_alive_count=self.tcp_keep_alive_count,
            tcp_keep_alive_interval=self.tcp_keep_alive_interval,
        )

    @cached_property
    def gcs_hook(self) -> GCSHook:
        """Create and return an GCSHook."""
        return GCSHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)

    def execute(self, context: Context):
        self.log.info("Calling HTTP method")
        response = self.http_hook.run(
            endpoint=self.endpoint, data=self.data, headers=self.headers, extra_options=self.extra_options
        )

        self.log.info("Uploading to GCS")
        self.gcs_hook.upload(
            data=response.content,
            bucket_name=self.bucket_name,
            object_name=self.object_name,
            mime_type=self.mime_type,
            gzip=self.gzip,
            encoding=self.encoding or response.encoding,
            chunk_size=self.chunk_size,
            timeout=self.timeout,
            num_max_attempts=self.num_max_attempts,
            metadata=self.metadata,
            cache_control=self.cache_control,
            user_project=self.user_project,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from tempfile import NamedTemporaryFile
from unittest import mock

import pytest

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.models import TaskInstance as TI
from airflow.providers.google.marketing_platform.operators.display_video import (
    GoogleDisplayVideo360CreateQueryOperator,
    GoogleDisplayVideo360CreateSDFDownloadTaskOperator,
    GoogleDisplayVideo360DeleteReportOperator,
    GoogleDisplayVideo360DownloadLineItemsOperator,
    GoogleDisplayVideo360DownloadReportV2Operator,
    GoogleDisplayVideo360RunQueryOperator,
    GoogleDisplayVideo360SDFtoGCSOperator,
    GoogleDisplayVideo360UploadLineItemsOperator,
)
from airflow.utils import timezone
from airflow.utils.session import create_session

API_VERSION = "v4"
GCP_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]

DEFAULT_DATE = timezone.datetime(2021, 1, 1)
REPORT_ID = "report_id"
BUCKET_NAME = "test_bucket"
REPORT_NAME = "test_report.csv"
QUERY_ID = FILENAME = "test.csv"
OBJECT_NAME = "object_name"
OPERATION_NAME = "test_operation"
RESOURCE_NAME = "resource/path/to/media"


class TestGoogleDisplayVideo360DeleteReportOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_execute(self, hook_mock):
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360DeleteReportOperator(
                report_id=QUERY_ID, api_version=API_VERSION, task_id="test_task"
            )
            op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=None,
        )
        hook_mock.return_value.delete_query.assert_called_once_with(query_id=QUERY_ID)


@pytest.mark.db_test
class TestGoogleDisplayVideo360DownloadReportV2Operator:
    def setup_method(self):
        with create_session() as session:
            session.query(TI).delete()

    def teardown_method(self):
        with create_session() as session:
            session.query(TI).delete()

    @pytest.mark.parametrize(
        "file_path, should_except", [("https://host/path", False), ("file:/path/to/file", True)]
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.shutil")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.urllib.request")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.tempfile")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.GCSHook")
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_execute(
        self,
        mock_hook,
        mock_gcs_hook,
        mock_temp,
        mock_request,
        mock_shutil,
        file_path,
        should_except,
    ):
        mock_temp.NamedTemporaryFile.return_value.__enter__.return_value.name = FILENAME
        mock_hook.return_value.get_report.return_value = {
            "metadata": {
                "status": {"state": "DONE"},
                "googleCloudStoragePath": file_path,
            }
        }
        # Create mock context with task_instance
        mock_context = {"task_instance": mock.Mock()}

        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360DownloadReportV2Operator(
                query_id=QUERY_ID,
                report_id=REPORT_ID,
                bucket_name=BUCKET_NAME,
                report_name=REPORT_NAME,
                task_id="test_task",
            )
            if should_except:
                with pytest.raises(AirflowException):
                    op.execute(context=mock_context)
                return
            op.execute(context=mock_context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v2",
            impersonation_chain=None,
        )
        mock_hook.return_value.get_report.assert_called_once_with(report_id=REPORT_ID, query_id=QUERY_ID)

        mock_gcs_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_gcs_hook.return_value.upload.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            filename=FILENAME,
            gzip=True,
            mime_type="text/csv",
            object_name=REPORT_NAME + ".gz",
        )
        mock_context["task_instance"].xcom_push.assert_called_once_with(
            key="report_name", value=REPORT_NAME + ".gz"
        )

    @pytest.mark.parametrize(
        "test_bucket_name",
        [BUCKET_NAME, f"gs://{BUCKET_NAME}", "XComArg", "{{ ti.xcom_pull(task_ids='taskflow_op') }}"],
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.shutil")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.urllib.request")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.tempfile")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.GCSHook")
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_set_bucket_name(
        self,
        mock_hook,
        mock_gcs_hook,
        mock_temp,
        mock_request,
        mock_shutil,
        test_bucket_name,
        dag_maker,
    ):
        mock_temp.NamedTemporaryFile.return_value.__enter__.return_value.name = FILENAME
        mock_hook.return_value.get_report.return_value = {
            "metadata": {"status": {"state": "DONE"}, "googleCloudStoragePath": "TEST"}
        }
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            with dag_maker(dag_id="test_set_bucket_name", start_date=DEFAULT_DATE) as dag:
                if BUCKET_NAME not in test_bucket_name:

                    @dag.task(task_id="taskflow_op")
                    def f():
                        return BUCKET_NAME

                    taskflow_op = f()

                GoogleDisplayVideo360DownloadReportV2Operator(
                    query_id=QUERY_ID,
                    report_id=REPORT_ID,
                    bucket_name=test_bucket_name if test_bucket_name != "XComArg" else taskflow_op,
                    report_name=REPORT_NAME,
                    task_id="test_task",
                )

            dr = dag_maker.create_dagrun()

            for ti in dr.get_task_instances():
                ti.run()

        mock_gcs_hook.return_value.upload.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            filename=FILENAME,
            gzip=True,
            mime_type="text/csv",
            object_name=REPORT_NAME + ".gz",
        )


class TestGoogleDisplayVideo360RunQueryOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_execute(self, hook_mock):
        parameters = {"param": "test"}

        # Create mock context with task_instance
        mock_context = {"task_instance": mock.Mock()}

        hook_mock.return_value.run_query.return_value = {
            "key": {
                "queryId": QUERY_ID,
                "reportId": REPORT_ID,
            }
        }
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360RunQueryOperator(
                query_id=QUERY_ID,
                parameters=parameters,
                api_version=API_VERSION,
                task_id="test_task",
            )
            op.execute(context=mock_context)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=None,
        )

        mock_context["task_instance"].xcom_push.assert_any_call(key="query_id", value=QUERY_ID)
        mock_context["task_instance"].xcom_push.assert_any_call(key="report_id", value=REPORT_ID)
        hook_mock.return_value.run_query.assert_called_once_with(query_id=QUERY_ID, params=parameters)


class TestGoogleDisplayVideo360DownloadLineItemsOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.GCSHook")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.tempfile")
    def test_execute(self, mock_temp, gcs_hook_mock, hook_mock):
        request_body = {
            "filterType": "filter_type",
            "filterIds": [],
            "format": "format",
            "fileSpec": "file_spec",
        }
        mock_temp.NamedTemporaryFile.return_value.__enter__.return_value.name = FILENAME
        gzip = False
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360DownloadLineItemsOperator(
                request_body=request_body,
                bucket_name=BUCKET_NAME,
                object_name=OBJECT_NAME,
                gzip=gzip,
                api_version=API_VERSION,
                gcp_conn_id=GCP_CONN_ID,
                task_id="test_task",
                impersonation_chain=IMPERSONATION_CHAIN,
            )
            op.execute(context=None)

        gcs_hook_mock.return_value.upload.assert_called_with(
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=FILENAME,
            gzip=gzip,
            mime_type="text/csv",
        )

        gcs_hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        hook_mock.return_value.download_line_items.assert_called_once_with(request_body=request_body)


class TestGoogleDisplayVideo360UploadLineItemsOperator:
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.tempfile")
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.GCSHook")
    def test_execute(self, gcs_hook_mock, hook_mock, mock_tempfile):
        line_items = "holy_hand_grenade"
        gcs_hook_mock.return_value.download.return_value = line_items
        mock_tempfile.NamedTemporaryFile.return_value.__enter__.return_value.name = FILENAME
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360UploadLineItemsOperator(
                bucket_name=BUCKET_NAME,
                object_name=OBJECT_NAME,
                api_version=API_VERSION,
                gcp_conn_id=GCP_CONN_ID,
                task_id="test_task",
            )
            op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=None,
        )

        gcs_hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )

        gcs_hook_mock.return_value.download.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=FILENAME,
        )
        hook_mock.return_value.upload_line_items.assert_called_once()
        hook_mock.return_value.upload_line_items.assert_called_once_with(line_items=line_items)


class TestGoogleDisplayVideo360SDFtoGCSOperator:
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.zipfile")
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.os")
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.tempfile.TemporaryDirectory"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.display_video.GCSHook")
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.open",
        new_callable=mock.mock_open,
    )
    def test_execute(self, mock_open, mock_hook, gcs_hook_mock, temp_dir_mock, os_mock, zipfile_mock):
        operation = {"response": {"resourceName": RESOURCE_NAME}}
        media = mock.Mock()

        mock_hook.return_value.get_sdf_download_operation.return_value = operation
        mock_hook.return_value.download_media.return_value = media

        tmp_dir = "/tmp/mock_dir"
        temp_dir_mock.return_value.__enter__.return_value = tmp_dir

        # Mock os behavior
        os_mock.path.join.side_effect = lambda *args: "/".join(args)
        os_mock.listdir.return_value = [FILENAME]

        # Mock zipfile behavior
        zipfile_mock.ZipFile.return_value.__enter__.return_value.extractall.return_value = None

        op = GoogleDisplayVideo360SDFtoGCSOperator(
            operation_name=OPERATION_NAME,
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            gzip=False,
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            task_id="test_task",
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        result = op.execute(context=None)

        # Assertions
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.get_sdf_download_operation.assert_called_once_with(
            operation_name=OPERATION_NAME
        )
        mock_hook.return_value.download_media.assert_called_once_with(resource_name=RESOURCE_NAME)
        mock_hook.return_value.download_content_from_request.assert_called_once()

        gcs_hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        gcs_hook_mock.return_value.upload.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=f"{tmp_dir}/{FILENAME}",
            gzip=False,
        )

        assert result == f"{BUCKET_NAME}/{OBJECT_NAME}"


class TestGoogleDisplayVideo360CreateSDFDownloadTaskOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_execute(self, mock_hook):
        body_request = {
            "version": "1",
            "id": "id",
            "filter": {"id": []},
        }
        test_name = "test_task"

        # Create mock context with task_instance
        mock_context = {"task_instance": mock.Mock()}

        mock_hook.return_value.create_sdf_download_operation.return_value = {"name": test_name}

        op = GoogleDisplayVideo360CreateSDFDownloadTaskOperator(
            body_request=body_request,
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            task_id="test_task",
        )

        op.execute(context=mock_context)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version=API_VERSION,
            impersonation_chain=None,
        )

        mock_hook.return_value.create_sdf_download_operation.assert_called_once()
        mock_hook.return_value.create_sdf_download_operation.assert_called_once_with(
            body_request=body_request
        )
        mock_context["task_instance"].xcom_push.assert_called_once_with(key="name", value=test_name)


class TestGoogleDisplayVideo360CreateQueryOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360Hook"
    )
    def test_execute(self, hook_mock):
        body = {"body": "test"}

        # Create mock context with task_instance
        mock_context = {"task_instance": mock.Mock()}

        hook_mock.return_value.create_query.return_value = {"queryId": QUERY_ID}
        with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
            op = GoogleDisplayVideo360CreateQueryOperator(body=body, task_id="test_task")
            op.execute(context=mock_context)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v2",
            impersonation_chain=None,
        )
        hook_mock.return_value.create_query.assert_called_once_with(query=body)
        mock_context["task_instance"].xcom_push.assert_called_once_with(key="query_id", value=QUERY_ID)

    def test_prepare_template(self):
        body = {"key": "value"}
        with NamedTemporaryFile("w+", suffix=".json") as f:
            f.write(json.dumps(body))
            f.flush()
            with pytest.warns(AirflowProviderDeprecationWarning):  # noqa: PT031
                op = GoogleDisplayVideo360CreateQueryOperator(body=body, task_id="test_task")
                op.prepare_template()

            assert isinstance(op.body, dict)
            assert op.body == body
